import datetime
from unittest.mock import MagicMock, patch

import numpy as np
import pandas as pd
import plotly.graph_objects as go
import pytest

from hdhelpers.exceptions import HelperException
from hdhelpers.plot_target_settings import (
    PlotTargetSettings,
    PlotTargetStyle,
    StatusColors,
)
from hdhelpers.user_functions import (
    get_and_pad_start_and_end_timestamp,
    get_colors_from_plot_target_settings,
    get_locale_from_plot_target_settings,
    get_y_axis_label,
    modify_timezone,
    plotly_fig_to_json_dict,
)


def test_get_y_axis_label_default():
    series = pd.Series()
    assert (
        get_y_axis_label(series=series, default_title="default_name", default_unit="default_unit")
        == "default_name [default_unit]"
    )


def test_get_y_axis_labeltitle_with_unit_metadata():
    series = pd.Series()
    series.attrs["single_metric_metadata"] = {
        "structured_metadata": {"metric": {"short_display_name": "name_from_metadata"}}
    }
    series.attrs["single_metric_metadata"]["structured_metadata"]["metric"]["unit"] = (
        "unit_from_metadata"
    )
    assert get_y_axis_label(series=series) == "name_from_metadata [unit_from_metadata]"


def test_get_no_colors_from_plot_target_settings():
    plot_target_settings_mock = MagicMock(return_value=PlotTargetSettings())
    with patch("hdhelpers.user_functions.get_plot_target_settings", plot_target_settings_mock):
        style_object = get_colors_from_plot_target_settings()
        assert isinstance(style_object, PlotTargetStyle)


def test_get_one_color_from_plot_target_settings():
    plot_target_settings_mock = MagicMock(
        return_value=PlotTargetSettings(
            plot_target_style=PlotTargetStyle(
                axes_label_color="#000000",
            )
        )
    )
    with patch("hdhelpers.user_functions.get_plot_target_settings", plot_target_settings_mock):
        style_object = get_colors_from_plot_target_settings()
        assert isinstance(style_object, PlotTargetStyle)


def test_get_all_colors_from_plot_target_settings():
    plot_target_settings_mock = MagicMock(
        return_value=PlotTargetSettings(
            plot_target_style=PlotTargetStyle(
                axes_label_color="#000000",
                background_color="#FFFFFF",
                grid_color="#8C8C98",
                line_colors=["#2FAE53", "#EB7C45", "#89CE6E", "#FFB058"],
                status_colors=StatusColors(
                    success_color="#2FAE53",
                    error_color="#EB6962",
                    warn_color="#9CE6E",
                    info_color="#80B0EC",
                ),
            )
        )
    )
    with patch("hdhelpers.user_functions.get_plot_target_settings", plot_target_settings_mock):
        style_object = get_colors_from_plot_target_settings()
        assert isinstance(style_object, PlotTargetStyle)


def test_get_no_locale_from_plot_target_settings():
    plot_target_settings_mock = MagicMock(return_value=PlotTargetSettings(plot_target_locale=None))
    with patch("hdhelpers.user_functions.get_plot_target_settings", plot_target_settings_mock):
        locale = get_locale_from_plot_target_settings()
        assert isinstance(locale, str | None)


def test_get_empty_locale_from_plot_target_settings():
    plot_target_settings_mock = MagicMock(return_value=PlotTargetSettings(plot_target_locale=""))
    with patch("hdhelpers.user_functions.get_plot_target_settings", plot_target_settings_mock):
        locale = get_locale_from_plot_target_settings()
        assert isinstance(locale, str | None)


def test_get_german_locale_from_plot_target_settings():
    plot_target_settings_mock = MagicMock(return_value=PlotTargetSettings(plot_target_locale="de"))
    with patch("hdhelpers.user_functions.get_plot_target_settings", plot_target_settings_mock):
        locale = get_locale_from_plot_target_settings()
        assert isinstance(locale, str | None)


@pytest.mark.parametrize(
    ("start", "end", "start_padding", "end_padding"),
    [
        ("2025-05-19T09:00:00+02:00", "2025-05-19T18:00:00+02:00", "1s", "1s"),
        ("2025-05-19T09:00:00+02:00", "2025-05-19T18:00:00+02:00", "1min", "1min"),
        ("2025-05-19T09:00:00+02:00", "2025-05-19T18:00:00+02:00", "1h", "1h"),
        ("2025-05-19T09:00:00+02:00", "2025-05-19T18:00:00+02:00", "1d", "1d"),
        ("2025-05-19T09:00:00+02:00", "2025-05-19T18:00:00+02:00", "1W", "1W"),
        ("2025-05-19T09:00:00+02:00", "2025-05-19T18:00:00+02:00", "1ME", "1ME"),
        ("2025-05-19T09:00:00+02:00", "2025-05-19T18:00:00+02:00", "1YE", "1YE"),
    ],
)
def test_get_and_pad_start_and_end_timestamp(start, end, start_padding, end_padding):
    start, end = get_and_pad_start_and_end_timestamp(
        pd.Series(), "Europe/Berlin", start, start_padding, end, end_padding
    )
    assert isinstance(start, pd.Timestamp)
    assert isinstance(end, pd.Timestamp)
    assert end >= start


def test_get_and_pad_none():
    with pytest.raises(HelperException):
        start, end = get_and_pad_start_and_end_timestamp(pd.Series())


@pytest.fixture()
def series_winter() -> pd.Series:
    winter = pd.Series(
        [0, 1, 2, 3],
        index=pd.to_datetime(
            ["2023-10-29 00:00", "2023-10-29 01:00", "2023-10-29 02:00", "2023-10-29 03:00"],
            format="%Y-%m-%d %H:%M",
            utc=True,
        ),
    )
    winter.attrs["foo"] = "bar"

    return winter


@pytest.fixture()
def series_summer() -> pd.Series:
    summer = pd.Series(
        [0, 1, 2, 3],
        index=pd.to_datetime(
            ["2023-03-25 23:00", "2023-03-26 00:00", "2023-03-26 01:00", "2023-03-26 02:00"],
            format="%Y-%m-%d %H:%M",
            utc=True,
        ),
    )
    summer.attrs["foo"] = "bar"
    return summer


@pytest.fixture()
def dataframe() -> pd.DataFrame:
    values = [1.0, 1.2, 1.2]
    timestamps = pd.to_datetime(
        [
            "2019-08-01T15:45:36.000Z",
            "2019-08-02T11:33:41.000Z",
            "2019-08-03T11:57:41.000Z",
        ],
        format="%Y-%m-%dT%H:%M:%S.%fZ",
    ).tz_localize("UTC")

    ts_df = pd.DataFrame({"timestamp": timestamps, "value": values})
    ts_df.attrs["foo"] = "bar"

    return ts_df


@pytest.fixture()
def multicolumn_frame() -> pd.DataFrame:
    values = [1.0, 1.2, 1.2]
    index = pd.to_datetime(
        [
            "2021-08-01T15:45:36.000Z",
            "2021-08-02T11:33:41.000Z",
            "2021-08-03T11:57:41.000Z",
        ],
        format="%Y-%m-%dT%H:%M:%S.%fZ",
    ).tz_localize("UTC")
    more_timestamps = pd.to_datetime(
        [
            "2020-08-01T15:45:36.000Z",
            "2020-08-02T11:33:41.000Z",
            "2020-08-03T11:57:41.000Z",
        ],
        format="%Y-%m-%dT%H:%M:%S.%fZ",
    ).tz_localize("UTC")
    timestamps = pd.to_datetime(
        [
            "2019-08-01T15:45:36.000Z",
            "2019-08-02T11:33:41.000Z",
            "2019-08-03T11:57:41.000Z",
        ],
        format="%Y-%m-%dT%H:%M:%S.%fZ",
    ).tz_localize("UTC")

    ts_df = pd.DataFrame(
        {"timestamp": timestamps, "values": values, "more_timestamps": more_timestamps}, index=index
    )
    ts_df.attrs["foo"] = "bar"

    return ts_df


def test_modify_timezone_good_dataframe(dataframe):
    local_summertime = modify_timezone(
        dataframe, to_timezone="Europe/Berlin", column_name="timestamp"
    )

    # German summer time starts in last Sunday in March at 2 am. --> UTC 1am
    timestamp_id = local_summertime.columns.get_loc("timestamp")
    assert local_summertime.iloc[1, timestamp_id].utcoffset() == datetime.timedelta(seconds=7200)
    assert local_summertime.iloc[2, timestamp_id].utcoffset() == datetime.timedelta(seconds=7200)
    assert "foo" in local_summertime.attrs


def test_modify_timezone_good_series(series_summer, series_winter):
    local_summertime = modify_timezone(series_summer, to_timezone="Europe/Berlin")
    local_wintertime = modify_timezone(series_winter, to_timezone="Europe/Berlin")
    cet = modify_timezone(series_winter, to_timezone="CET")
    utc_wintertime = modify_timezone(local_wintertime, to_timezone="UTC")

    # German summer time starts in last Sunday in March at 2 am. --> UTC 1am
    assert local_summertime.index[1].utcoffset() == datetime.timedelta(seconds=3600)
    assert local_summertime.index[2].utcoffset() == datetime.timedelta(seconds=7200)
    assert "foo" in local_summertime.attrs

    # German winter time starts in last Sunday in October at 3 am. --> UTC: 1am
    assert local_wintertime.index[0].utcoffset() == datetime.timedelta(seconds=7200)
    assert local_wintertime.index[1].utcoffset() == datetime.timedelta(seconds=3600)
    assert "foo" in local_wintertime.attrs

    # cet is equal to German winter time
    assert local_wintertime.index[1] == cet.index[1]

    # reversing works
    pd.testing.assert_series_equal(series_winter, utc_wintertime)

    # timedelta not influneced by new timezone
    np.testing.assert_array_equal(
        pd.to_timedelta(local_summertime.index[1:] - local_summertime.index[:-1])
        .total_seconds()
        .values,
        [3600.0, 3600.0, 3600.0],
    )


def test_modify_timezone_wrong_tzname(series_summer):
    with pytest.raises(ValueError, match="Timezone not known*"):
        _ = modify_timezone(series_summer, to_timezone="Europe/Berlin2")


def test_named_series(series_summer):
    data = pd.Series(series_summer.index)
    data.name = "timestamp"
    data.attrs = series_summer.attrs
    modified_data = modify_timezone(data, to_timezone="Europe/Berlin", column_name="timestamp")
    assert modified_data[1].utcoffset() == datetime.timedelta(seconds=3600)
    assert "foo" in modified_data.attrs


def test_named_series_using_index(series_summer):
    data = series_summer
    data.name = "timestamp"
    modified_data = modify_timezone(data, to_timezone="Europe/Berlin", column_name=None)
    assert modified_data.index[0].utcoffset() == datetime.timedelta(seconds=3600)
    assert "foo" in modified_data.attrs


def test_column_not_known(series_summer, dataframe):
    data = pd.Series(series_summer.index)
    data.name = "timestamp"

    with pytest.raises(KeyError, match="Column name*"):
        _ = modify_timezone(data, to_timezone="Europe/Berlin", column_name="timestamp2")

    with pytest.raises(KeyError, match="Column name*"):
        _ = modify_timezone(dataframe, to_timezone="Europe/Berlin", column_name="timestamp2")


def test_modify_timezone_no_tz_known(series_summer):
    series_summer.index = series_summer.index.tz_localize(None)
    with pytest.raises(TypeError, match="Entries to convert do not contain valid timestamps*"):
        _ = modify_timezone(series_summer, to_timezone="Europe/Berlin")


def test_modify_timezone_multicolumn_dataframe(multicolumn_frame):
    local_summertime = modify_timezone(
        multicolumn_frame,
        to_timezone="Europe/Berlin",
        column_names=["timestamp", "more_timestamps"],
    )

    # German summer time starts in last Sunday in March at 2 am. --> UTC 1am
    timestamp_id = local_summertime.columns.get_loc("timestamp")
    timestamp_id_2 = local_summertime.columns.get_loc("more_timestamps")
    assert local_summertime.iloc[1, timestamp_id].utcoffset() == datetime.timedelta(seconds=7200)
    assert local_summertime.iloc[1, timestamp_id_2].utcoffset() == datetime.timedelta(seconds=7200)
    assert local_summertime.index[1].utcoffset() == datetime.timedelta(seconds=7200)
    assert "foo" in local_summertime.attrs


def test_modify_timezone_multicolumn_dataframe_without_index(multicolumn_frame):
    local_summertime = modify_timezone(
        multicolumn_frame,
        to_timezone="Europe/Berlin",
        column_names=["timestamp", "more_timestamps"],
        convert_index=False,
    )

    # German summer time starts in last Sunday in March at 2 am. --> UTC 1am
    timestamp_id = local_summertime.columns.get_loc("timestamp")
    timestamp_id_2 = local_summertime.columns.get_loc("more_timestamps")
    assert local_summertime.iloc[1, timestamp_id].utcoffset() == datetime.timedelta(seconds=7200)
    assert local_summertime.iloc[1, timestamp_id_2].utcoffset() == datetime.timedelta(seconds=7200)
    assert local_summertime.index[1].utcoffset() == datetime.timedelta(seconds=0)
    assert "foo" in local_summertime.attrs


def test_plot_target_timezone(series_summer):
    plot_target_settings_mock = MagicMock(
        return_value=PlotTargetSettings(plot_target_timezone="Europe/Berlin")
    )
    with patch("hdhelpers.user_functions.get_plot_target_settings", plot_target_settings_mock):
        modified_data = modify_timezone(series_summer)
        assert modified_data.index[1].utcoffset() == datetime.timedelta(seconds=3600)


def test_modify_timestamp():
    modified_timestamp = modify_timezone(
        pd.to_datetime("2023-03-25 23:00", utc=True), to_timezone="Europe/Berlin"
    )
    assert modified_timestamp.utcoffset() == datetime.timedelta(seconds=3600)


def test_plotly_fig_to_json_dict_defaults():
    plotly_fig = go.Figure()
    plotly_fig.add_trace(
        go.Scatter(
            x=[1, 2, 3],
            y=[9, 8, 7],
            name="Foo",
        )
    )
    json_dict = plotly_fig_to_json_dict(plotly_fig)
    assert len(json_dict.get("layout", {}).get("template", {}).get("layout", {})["colorway"]) > 0
    assert json_dict.get("layout", {}).get("margin", {})["autoexpand"]
    assert json_dict.get("layout", {}).get("margin", {})["l"] == 0
    assert json_dict.get("layout", {}).get("margin", {})["r"] == 0
    assert json_dict.get("layout", {}).get("margin", {})["b"] == 0
    assert json_dict.get("layout", {}).get("margin", {})["t"] == 0
    assert json_dict.get("layout", {}).get("margin", {})["pad"] == 0
    assert not json_dict.get("config", {})["displaylogo"]
    assert not json_dict.get("config", {})["displayModeBar"]


def test_plotly_fig_to_json_dict_set_everything():
    plotly_fig = go.Figure()
    plotly_fig.add_trace(
        go.Scatter(
            x=[1, 2, 3],
            y=[9, 8, 7],
            name="Foo",
        )
    )
    json_dict = plotly_fig_to_json_dict(
        fig=plotly_fig,
        add_config_settings=False,
        hide_legend=True,
        hide_x_title=True,
        remove_plotly_bar=False,
        remove_plotly_icon=False,
        update_x_axes_tickformat=True,
        use_default_standoff=True,
        use_minimum_margin=False,
        use_muplot_axes_color=True,
        use_muplot_grid=True,
        use_muplot_line_and_markers=True,
        use_platform_background=True,
        use_platform_defaults=True,
        use_simple_white_template=False,
    )
    assert isinstance(json_dict, dict)

    assert len(json_dict.get("layout", {}).get("template", {}).get("layout", {})["colorway"]) > 0
    assert json_dict.get("layout", {}).get("margin", {}) == {}
    assert "displaylogo" not in json_dict.get("config", {})
    assert "displayModeBar" not in json_dict.get("config", {})



import datetime
import logging

from pydantic import BaseModel, Field

logger = logging.getLogger(__name__)


class StatusColors(BaseModel):
    """Collection of status-related colors

    Unlike the other colors in PlotTargetSettings, these colors do not have a predefined use.
    Instead they should be used contextually, e.g. when displaying the sensor status of an asset.
    """

    success_color: str | None = Field(
        None, description="Color of markers that signal success as a hexcode"
    )
    error_color: str | None = Field(
        None, description="Color of markers that signal errors as a hexcode"
    )
    warn_color: str | None = Field(
        None, description="Color of markers that signal warnings as a hexcode"
    )
    info_color: str | None = Field(
        None, description="Color of markers that signal informativeness as a hexcode"
    )


class PlotTargetStyle(BaseModel):
    axes_label_color: str | None = Field(
        None, description="Color of the tick labels of all axes as a hex code"
    )
    background_color: str | None = Field(
        None, description="Color of the panel background as a hex code"
    )
    grid_color: str | None = Field(
        None, description="Color of the grid as a hex code that may be drawn into the background"
    )
    line_colors: list[str] | None = Field(
        None,
        description="""List of colors to be used for plot traces.
            Will be set as colorway by plotly_fig_to_json_dict,
            so the colors are only applied where no explicit trace color is set""",
    )
    status_colors: StatusColors = Field(
        StatusColors(),  # type: ignore
        description="Has the properties success_color, error_color, warn_color, info_color",
    )


class PlotTargetSettings(BaseModel):
    """Settings that plot components can/should use

    Some Plotly settings like locale or the timezone of timestamps must be set
    by Python and cannot easily be set by plotly.js in a frontend.

    They can be provided to execution endpoints as part of the ExecByIdBase payload,
    are made accessible to components using the execution context.

    hdhelpers provides helper functions to access them at runtime.
    """

    plot_target_timezone: str | None = Field(
        None,
        description="""The timezone plot components should use for datetime axes etc.
             Usually via
             s.index=pd.to_datetime(s.index, utc=True).tz_convert(plot_target_timezone)""",
        examples=["Europe/Berlin"],
    )
    plot_target_locale: str | None = Field(
        None,
        description="""Locale to set for plots, e.g. to write weekdays in the user's language.
             This has to be set in the config of the plotly figure dict and the plotly.js
             must have the associated plotly local scripts loaded.""",
    )
    plot_target_style: PlotTargetStyle = Field(
        PlotTargetStyle(),  # type: ignore
        description="Colors to use in the plot",
    )
    datetime_tick_format: str | None = Field(
        None, description="Tickformat to use for datetime axes", examples=["%H:%M<br>%d.%m.%Y"]
    )
    datetime_x_axes_range_start: datetime.datetime | None = Field(
        None, description="datetime range start which plots should set as x axis range"
    )

    datetime_x_axes_range_end: datetime.datetime | None = Field(
        None, description="datetime range end which plots should set as x axis range"
    )


def get_plot_target_settings() -> PlotTargetSettings:
    """Obtain plot settings from runtime execution context.

    If hetdesrun is not importable or this context field is not set,
    return default values.
    """
    try:
        from hetdesrun.runtime.context import (  # type: ignore  # noqa: PLC0415
            get_runtime_exec_context,
        )

        plot_target_settings = get_runtime_exec_context().plot_target_settings
        if not isinstance(plot_target_settings, PlotTargetSettings):
            raise TypeError("plot_target_settings must be instance of PlotTargetSettings")

        return plot_target_settings
    except (ImportError, TypeError):
        logger.warning("Could not load runtime exec context, import failed! Switch to defaults.")
        # return defaults if hetdesrun is not available as import
        return PlotTargetSettings()  # type: ignore



import logging
from datetime import datetime
from typing import Any

import pandas as pd
from pandas.tseries.frequencies import to_offset

from hdhelpers.exceptions import HelperException
from hdhelpers.plot_target_settings import get_plot_target_settings

logger = logging.getLogger(__name__)


def _convert_to_optional_timezone(object_to_convert: Any, to_timezone: str | None) -> Any:
    """Convert object_to_convert to to_timezone if not None,
    or to its own timezone if aware
    or to UTC otherwise"""
    if to_timezone is None:
        if object_to_convert.tz is None:
            return object_to_convert.tz_localize("UTC")
        return object_to_convert
    if object_to_convert.tz is None:
        return object_to_convert.tz_localize(to_timezone)
    return object_to_convert.tz_convert(to_timezone)


def _get_display_name(series: pd.Series, default_title: str = "") -> str:
    """Get name for y-axis label from metadata

    Tries to get the name from series.attrs according to the conventions of thehetida platform.
    If such metadata doesn't exist, the default_title is returned instead.
    """
    try:
        title = (
            series.attrs.get("single_metric_metadata", {})
            .get("structured_metadata", {})
            .get("metric", {})["short_display_name"]
        )
        if not isinstance(title, str):
            raise HelperException("Expected short_display_name to be a string, but it is not!")
    except (KeyError, HelperException) as exc:
        msg = (
            'Expected attrs["single_metric_metadata"]["structured_metadata"]["metric"]',
            '["short_display_name"] but got incorrect keys',
        )
        logger.warning(msg=msg, exc_info=exc)
        title = default_title
    return title


def _get_unit(series: pd.Series, default_unit: str = "") -> str:
    """Get unit for y-axis label from metadata

    Tries to get the unit from series.attrs according to the conventions of thehetida platform.
    If such metadata doesn't exist, the default_unit is returned instead.
    """
    try:
        unit = (
            series.attrs.get("single_metric_metadata", {})
            .get("structured_metadata", {})
            .get("metric", {})["unit"]
        )
        if not isinstance(unit, str):
            raise HelperException("Expected unit to be a string, but it is not!")
    except (KeyError, HelperException) as exc:
        msg = 'Expected attrs["single_metric_metadata"]["structured_metadata"]["metric"]["unit"'
        "] but got incorrect keys"
        logger.warning(msg=msg, exc_info=exc)
        unit = default_unit
    return unit


def _pad_start(timestamp: pd.Timestamp, padding: str | None) -> pd.Timestamp:
    """Subtracts padding from the timestamp

    That padding has to be formatted to be compatible with pandas.tseries.frequencies.to_offset().
    """
    if padding is None:
        return timestamp
    try:
        return timestamp - to_offset(padding)
    except ValueError as exc:
        raise HelperException(
            f"{padding} as padding value is an invalid frequency. "
            "Use something compatible with pandas.tseries.frequencies.to_offset()"
        ) from exc


def _pad_end(timestamp: pd.Timestamp, padding: str | None) -> pd.Timestamp:
    """Adds padding to the timestamp

    That padding has to be formatted to be compatible with pandas.tseries.frequencies.to_offset().
    """
    if padding is None:
        return timestamp
    try:
        return timestamp + to_offset(padding)
    except ValueError as exc:
        raise HelperException(
            f"{padding} as padding value is an invalid frequency. "
            "Use something compatible with pandas.tseries.frequencies.to_offset()"
        ) from exc


def _get_start_timestamp(
    series: pd.Series, timestamp: datetime | str | None
) -> pd.Timestamp | None:
    """Get the start timestamp  hierarchically

    Will check for an explicit input timestamp first, then check PlotTargetSettings, then the series
    metadata, and if both are None, will take the first series entry as start timestamp.
    If the series is also empty, None is returned.
    """
    if timestamp is not None:
        return _to_datetime(timestamp)

    plot_target_settings = get_plot_target_settings()

    timestamp = plot_target_settings.datetime_x_axes_range_start

    if timestamp is None:
        key = "ref_interval_start_timestamp"
        try:
            timestamp = series.attrs.get("single_metric_dataset_metadata", {})[key]
        except KeyError as exc:
            msg = f"""Expected key structure not found:
             attrs["single_metric_dataset_metadata"]["{key}"]"""
            logger.warning(msg=msg, exc_info=exc)
            if len(series) > 0:
                timestamp = series.index[0]

    return _to_datetime(timestamp)


def _get_end_timestamp(series: pd.Series, timestamp: datetime | str | None) -> pd.Timestamp | None:
    """Get the end timestamp hierarchically

    Will check for an explicit input timestamp first, then check PlotTargetSettings, then the series
    metadata, and if both are None, will take the last series entry as end timestamp.
    If the series is also empty, None is returned.
    """
    if timestamp is not None:
        return _to_datetime(timestamp)

    plot_target_settings = get_plot_target_settings()

    timestamp = plot_target_settings.datetime_x_axes_range_end

    if timestamp is None:
        key = "ref_interval_end_timestamp"
        try:
            timestamp = series.attrs.get("single_metric_dataset_metadata", {})[key]
        except KeyError as exc:
            msg = f"""Expected key structure not found:
             attrs["single_metric_dataset_metadata"]["{key}"]"""
            logger.warning(msg=msg, exc_info=exc)
            if len(series) > 0:
                timestamp = series.index[-1]

    return _to_datetime(timestamp)


def _to_datetime(timestamp: datetime | str | int | None) -> pd.Timestamp | None:
    """Turn datetime string or integer into a pandas timestamp

    Integer values are interpreted as epoch in seconds.
    String values are accepted in any format compatible with pd.to_datetime
    and interpreted in seconds.
    The timezone is set to utc in both cases, other timezones can be set via modify_timezone."""
    if timestamp is None:
        return None
    if isinstance(timestamp, int):
        timestamp = pd.to_datetime(timestamp, unit="s", utc=True)
    elif isinstance(timestamp, str | datetime):
        timestamp = pd.to_datetime(timestamp, utc=True)
    else:
        raise HelperException("Unexpected timestamp type, please use str or int!")
    return timestamp



from hdhelpers.exceptions import ComponentException, HelperException, InsufficientPlottingData
from hdhelpers.helper_functions import (
    _get_display_name,
    _get_end_timestamp,
    _get_start_timestamp,
    _get_unit,
    _pad_end,
    _pad_start,
    _to_datetime,
)
from hdhelpers.plot_target_settings import (
    PlotTargetSettings,
    PlotTargetStyle,
    StatusColors,
    get_plot_target_settings,
)
from hdhelpers.user_functions import (
    get_and_pad_start_and_end_timestamp,
    get_colors_from_plot_target_settings,
    get_locale_from_plot_target_settings,
    get_y_axis_label,
    modify_timezone,
    plotly_fig_to_json_dict,
)

__all__ = [
    "ComponentException",
    "HelperException",
    "InsufficientPlottingData",
    "PlotTargetSettings",
    "PlotTargetStyle",
    "StatusColors",
    "_get_display_name",
    "_get_end_timestamp",
    "_get_start_timestamp",
    "_get_unit",
    "_pad_end",
    "_pad_start",
    "_to_datetime",
    "get_and_pad_start_and_end_timestamp",
    "get_colors_from_plot_target_settings",
    "get_locale_from_plot_target_settings",
    "get_plot_target_settings",
    "get_y_axis_label",
    "modify_timezone",
    "plotly_fig_to_json_dict",
]



import json
import logging
from datetime import datetime
from typing import Any
from warnings import warn

import pandas as pd
import pytz
from plotly.graph_objects import Figure  # type: ignore
from plotly.utils import PlotlyJSONEncoder  # type: ignore

from hdhelpers.exceptions import HelperException
from hdhelpers.helper_functions import (
    _convert_to_optional_timezone,
    _get_display_name,
    _get_end_timestamp,
    _get_start_timestamp,
    _get_unit,
    _pad_end,
    _pad_start,
)
from hdhelpers.plot_target_settings import PlotTargetStyle, get_plot_target_settings

logger = logging.getLogger(__name__)


def get_colors_from_plot_target_settings() -> PlotTargetStyle:
    """Get thematically coherent colors for customizing plots

    Most color uses are already covered by the default settings of plotly_fig_to_json_dict().
    They are still included here in case coloring other plot elements in the same color is desired.
    Each color is given as a hex code, line_colors is a list of such, as specified in
    PlotTargetStyle.
    """
    plot_target_settings = get_plot_target_settings()

    return plot_target_settings.plot_target_style


def get_locale_from_plot_target_settings() -> str | None:
    """Get language for customizing text elements in plots

    Axis ticks are already covered by the default settings of plotly_fig_to_json_dict().
    The language of custom text elements should be adjusted to the locale.
    """
    plot_target_settings = get_plot_target_settings()

    return plot_target_settings.plot_target_locale


def get_y_axis_label(series: pd.Series, default_title: str = "", default_unit: str = "") -> str:
    """Get full y-axis label from metadata

    Combines the title and unit provided by _get_display_name and _get_unit.
    """
    title = _get_display_name(series, default_title)
    unit = _get_unit(series, default_unit)
    if len(unit) > 0:
        title = f"{title} [{unit}]"
    return title


def get_and_pad_start_and_end_timestamp(
    series: pd.Series,
    timezone: str | None = None,
    start: datetime | str | None = None,
    start_padding: str | None = None,
    end: datetime | str | None = None,
    end_padding: str | None = None,
) -> tuple[pd.Timestamp, pd.Timestamp]:
    """Get time period displayed on the x-axis

    Retrieves the start and end timestamps, prioritizing the explicit "start" and "end" parameters
    over the metadata of "series" and using the first and last index of the series if neither is
    given. If a padding is given, the respective timestamp is adjusted. That padding has to be
    formatted to be compatible with pandas.tseries.frequencies.to_offset().
    """
    # Get start and end
    start = _get_start_timestamp(series, start)
    end = _get_end_timestamp(series, end)

    if start is None:
        raise HelperException("No start timestamp found!")
    start_timestamp = start
    if end is None:
        raise HelperException("No end timestamp found!")
    end_timestamp = end

    # Convert timezone
    if timezone is not None:
        start_with_timezone = modify_timezone(start_timestamp, timezone)
        end_with_timezone = modify_timezone(end_timestamp, timezone)
    else:
        start_with_timezone = start_timestamp
        end_with_timezone = end_timestamp

    # Optionally add padding
    start_padded = _pad_start(start_with_timezone, start_padding)
    end_padded = _pad_end(end_with_timezone, end_padding)

    return start_padded, end_padded


def modify_timezone[T: (pd.Timestamp, pd.Series, pd.DataFrame)](  # noqa: PLR0912
    object_to_convert: T,
    to_timezone: str | None = None,
    column_name: str | None = None,
    column_names: list[str] | None = None,
    convert_index: bool = True,
) -> T:
    """Modifies timestamps to a certain timezone

    Keyword arguments:
    object_to_convert -- pd.Timestamp, pd.Series or pd.DataFrame where timezone is modified
    to_timezone -- timezone to convert to, e.g. for German time use Europe/Berlin.
    See possible timezone strings in pandas tz_convert method or pytz all_timezones list.
    column_name -- column_name to apply, default is index as pd.Series have timestamps in index
    """
    if not isinstance(object_to_convert, pd.Timestamp | pd.Series | pd.DataFrame):
        raise TypeError(
            f"object_to_convert is {type(object_to_convert)} not pd.Series | pd.DataFrame"
        )
    if column_names is None:
        column_names = []

    try:
        if to_timezone is None:
            plot_target_settings = get_plot_target_settings()
            if plot_target_settings.plot_target_timezone is not None:
                to_timezone = plot_target_settings.plot_target_timezone

        if isinstance(object_to_convert, pd.Timestamp):
            return _convert_to_optional_timezone(object_to_convert, to_timezone)

        if isinstance(object_to_convert, pd.Series):
            new_object = object_to_convert.to_frame(name=object_to_convert.name)
        else:
            new_object = object_to_convert.copy(deep=True)

        # Both column_name branches exist purely for backwards compatibility,
        # only convert_index should stay.
        if column_name is None and convert_index:
            new_object.index = _convert_to_optional_timezone(
                pd.to_datetime(new_object.index), to_timezone
            )
        if column_name is not None:
            warn(
                """The parameter 'column_name' will soon be deprecated in favor of
                the more flexible 'columns_names'""",
                DeprecationWarning,
                stacklevel=2,
            )
            new_object[column_name] = _convert_to_optional_timezone(
                pd.to_datetime(new_object[column_name]).dt, to_timezone
            )
            column_names.append(column_name)

        if len(column_names) == 0:
            if isinstance(object_to_convert, pd.Series):
                new_object.index = _convert_to_optional_timezone(
                    pd.to_datetime(new_object.index), to_timezone
                )
                msg = f"Converted index to datetime starting with {object_to_convert.index[0]}"
                logger.debug(msg=msg)
            elif isinstance(new_object, pd.DataFrame) and "timestamp" in new_object.columns:
                new_object["timestamp"] = _convert_to_optional_timezone(
                    pd.to_datetime(new_object["timestamp"]).dt, to_timezone
                )
                msg = f"""Converted column "timestamp" to datetime starting with
                {object_to_convert["timestamp"][0]}"""
                logger.debug(msg=msg)
        if len(column_names) > 0:
            for column in column_names:
                new_object[column] = _convert_to_optional_timezone(
                    pd.to_datetime(new_object[column]).dt, to_timezone
                )

        if not isinstance(object_to_convert, pd.Series):
            new_object.attrs = object_to_convert.attrs
            return new_object

        series_object = pd.Series(
            new_object[object_to_convert.name],
            index=new_object.index,
            name=object_to_convert.name,
        )
        series_object.attrs = object_to_convert.attrs

        return series_object

    except pytz.exceptions.UnknownTimeZoneError as exc:
        possible_timezone = pytz.all_timezones
        raise ValueError(f"""Timezone not known, please choose from {possible_timezone}""") from exc
    except (AttributeError, pytz.exceptions.NonExistentTimeError) as exc:
        raise TypeError("Entries to convert do not contain valid timestamps") from exc
    except KeyError as exc:
        exc.add_note(f"Column name {column_name} not in object_to_convert")
        raise


def plotly_fig_to_json_dict(  # noqa: PLR0912, PLR0915
    fig: Figure,
    add_config_settings: bool = True,
    hide_legend: bool | None = None,
    hide_x_title: bool | None = None,
    remove_plotly_bar: bool | None = None,
    remove_plotly_icon: bool = True,
    update_x_axes_tickformat: bool | None = None,
    use_default_standoff: bool = False,
    use_minimum_margin: bool = True,
    use_muplot_axes_color: bool | None = None,
    use_muplot_grid: bool | None = None,
    use_muplot_line_and_markers: bool | None = None,
    use_platform_background: bool | None = None,
    use_platform_colorway: bool = True,
    use_platform_defaults: bool = True,
    use_simple_white_template: bool = True,
) -> Any:
    """Turn Plotly figure into a Python dict-like object

    This function can be used in visualization components to obtain the
    correct plotly json-like object from a Plotly Figure object.

    Additionally, this function has a dozen boolean parameters that can be
    set to standardize certain aspects of the plot styling in accordance
    with the hetida platform.

    See visualization components from the accompanying base components for
    examples on usage.
    """
    if use_platform_defaults:
        if hide_legend is None:
            hide_legend = True
        if hide_x_title is None:
            hide_x_title = True
        if remove_plotly_bar is None:
            remove_plotly_bar = True
        if update_x_axes_tickformat is None:
            update_x_axes_tickformat = True
        if use_default_standoff is None:
            use_default_standoff = True
        if use_muplot_axes_color is None:
            use_muplot_axes_color = True
        if use_muplot_grid is None:
            use_muplot_grid = True
        if use_muplot_line_and_markers is None:
            use_muplot_line_and_markers = True
        if use_platform_background is None:
            use_platform_background = True
    else:
        if hide_legend is None:
            hide_legend = False
        if hide_x_title is None:
            hide_x_title = False
        if remove_plotly_bar is None:
            remove_plotly_bar = False
        if update_x_axes_tickformat is None:
            update_x_axes_tickformat = False
        if use_default_standoff is None:
            use_default_standoff = False
        if use_muplot_axes_color is None:
            use_muplot_axes_color = False
        if use_muplot_grid is None:
            use_muplot_grid = False
        if use_muplot_line_and_markers is None:
            use_muplot_line_and_markers = False
        if use_platform_background is None:
            use_platform_background = False

    plot_target_settings = get_plot_target_settings()

    if use_platform_colorway and plot_target_settings.plot_target_style.line_colors is not None:
        fig.update_layout(colorway=plot_target_settings.plot_target_style.line_colors)

    if use_simple_white_template:
        fig.update_layout({"template": "simple_white"})

    if (
        use_platform_background
        and plot_target_settings.plot_target_style.background_color is not None
    ):
        fig.update_layout(
            {
                "paper_bgcolor": plot_target_settings.plot_target_style.background_color,
                "plot_bgcolor": "rgba(0,0,0,0)",
            }
        )

    if hide_legend:
        fig.update_layout(showlegend=False)

    if hide_x_title:
        fig.update_xaxes(title_text="")

    if update_x_axes_tickformat and plot_target_settings.datetime_tick_format is not None:
        fig.update_xaxes(tickformat=plot_target_settings.datetime_tick_format)

    if (
        use_muplot_axes_color
        and plot_target_settings.plot_target_style.axes_label_color is not None
    ):
        fig.update_xaxes(color=plot_target_settings.plot_target_style.axes_label_color)
        fig.update_yaxes(color=plot_target_settings.plot_target_style.axes_label_color)

    if use_default_standoff:
        fig.update_yaxes(title_standoff=5)

    if use_muplot_line_and_markers:
        fig.update_traces(
            {
                "marker": {"size": 3},
                "line": {"width": 1},
                "mode": "lines+markers",
                "marker_symbol": "circle",
            }
        )

    if use_minimum_margin:
        fig.update_layout(
            {"margin": {"autoexpand": True, "l": 0, "r": 0, "b": 0, "t": 0, "pad": 0}}
        )

    if use_muplot_grid and plot_target_settings.plot_target_style.grid_color is not None:
        grid_dict = {
            "showgrid": True,
            "gridcolor": plot_target_settings.plot_target_style.grid_color,
            "zeroline": True,
            "zerolinecolor": plot_target_settings.plot_target_style.grid_color,
        }
        fig.update_layout({"xaxis": grid_dict, "yaxis": grid_dict})

    fig_dict_obj = fig.to_plotly_json()
    if not "config" in fig_dict_obj:
        fig_dict_obj["config"] = {}

    if add_config_settings and plot_target_settings.plot_target_locale is not None:
        fig_dict_obj["config"]["locale"] = plot_target_settings.plot_target_locale

    if remove_plotly_bar:
        fig_dict_obj["config"]["displayModeBar"] = False

    if remove_plotly_icon:
        fig_dict_obj["config"]["displaylogo"] = False

    # possibly quite inefficient (multiple serialisation / deserialization) but
    # guarantees that the PlotlyJSONEncoder is used and so the resulting Json
    # should be definitely compatible with the plotly javascript library:
    return json.loads(json.dumps(fig_dict_obj, cls=PlotlyJSONEncoder))



from typing import Any


class ComponentException(Exception):
    """Exception to re-raise exceptions with error code raised in the component code."""

    __is_hetida_designer_exception__ = True

    def __init__(
        self,
        *args: Any,
        error_code: int | str = "",
        extra_information: dict | None = None,
        **kwargs: Any,
    ) -> None:
        if not isinstance(error_code, int | str):
            raise ValueError("The ComponentException.error_code must be int or string!")
        self.error_code = error_code
        self.extra_information = extra_information
        super().__init__(*args, **kwargs)


class HelperException(Exception):
    """Exception to re-raise exceptions with error code raised in the code of the hdhelpers
    package."""

    __is_hetida_designer_exception__ = True

    def __init__(
        self,
        *args: Any,
        error_code: int | str = "",
        extra_information: dict | None = None,
        **kwargs: Any,
    ) -> None:
        if not isinstance(error_code, int | str):
            raise ValueError("The HelperException.error_code must be int or string!")
        self.error_code = error_code
        self.extra_information = extra_information
        super().__init__(*args, **kwargs)


class InsufficientPlottingData(ComponentException):
    """A plot component has insufficient data to generate a meaningful plot

    This exception class should be used when custom plots generated by hetida
    designer are integrated in other frontends. This allows the frontend to
    handle the case of e.g. no data to show in a sensible way, surpressing an
    empty plot and showing an adequate message instead.
    """



import datetime
from unittest.mock import MagicMock, patch

import pandas as pd
import pytest

from hdhelpers.exceptions import HelperException
from hdhelpers.helper_functions import (
    _convert_to_optional_timezone,
    _get_display_name,
    _get_end_timestamp,
    _get_start_timestamp,
    _get_unit,
    _pad_end,
    _pad_start,
    _to_datetime,
)
from hdhelpers.plot_target_settings import PlotTargetSettings


def test_convert_to_optional_timezone_naive_none():
    assert (
        _convert_to_optional_timezone(pd.to_datetime("2025-01-01T01:00:00"), None).tz
        == datetime.timezone.utc
    )


def test_convert_to_optional_timezone_aware_none():
    assert _convert_to_optional_timezone(
        pd.to_datetime("2025-01-01T01:00:00+05:00"), None
    ).tz == datetime.timezone(datetime.timedelta(seconds=18000))


def test_convert_to_optional_timezone_naive_given():
    timestamp = _convert_to_optional_timezone(
        pd.to_datetime("2025-01-01T01:00:00"), "Europe/Berlin"
    )
    assert timestamp.utcoffset() == datetime.timedelta(seconds=3600)


def test_convert_to_optional_timezone_aware_given():
    timestamp = _convert_to_optional_timezone(
        pd.to_datetime("2025-01-01T01:00:00+05:00"), "Europe/Berlin"
    )
    assert timestamp.utcoffset() == datetime.timedelta(seconds=3600)


def test_get_display_name_default():
    series = pd.Series()
    assert _get_display_name(series=series, default_title="default_name") == "default_name"


def test_get_display_name_metadata():
    series = pd.Series()
    series.attrs["single_metric_metadata"] = {
        "structured_metadata": {"metric": {"short_display_name": "name_from_metadata"}}
    }
    assert _get_display_name(series=series) == "name_from_metadata"


def test_get_unit_default():
    series = pd.Series()
    assert _get_unit(series=series, default_unit="default_unit") == "default_unit"


def test_get_unit_metadata():
    series = pd.Series()
    series.attrs["single_metric_metadata"] = {
        "structured_metadata": {"metric": {"unit": "unit_from_metadata"}}
    }
    assert _get_unit(series=series) == "unit_from_metadata"


def test_get_start_timestamp_directly():
    timestamp = _get_start_timestamp(pd.Series(), "2025-05-28T09:00:00+02:00")
    assert isinstance(timestamp, pd.Timestamp)


def test_get_start_timestamp_attrs():
    series = pd.Series()
    series.attrs = {
        "single_metric_dataset_metadata": {
            "ref_interval_start_timestamp": "2025-05-28T09:00:00+02:00"
        }
    }
    timestamp = _get_start_timestamp(series, None)
    assert isinstance(timestamp, pd.Timestamp)


def test_get_start_timestamp_plot_target_settings():
    plot_target_settings_mock = MagicMock(
        return_value=PlotTargetSettings(datetime_x_axes_range_start="2025-05-28T09:00:00+02:00")
    )
    with patch("hdhelpers.helper_functions.get_plot_target_settings", plot_target_settings_mock):
        timestamp = _get_start_timestamp(pd.Series(), None)
        assert isinstance(timestamp, pd.Timestamp)


def test_get_end_timestamp_directly():
    timestamp = _get_end_timestamp(pd.Series(), "2025-05-28T18:00:00+02:00")
    assert isinstance(timestamp, pd.Timestamp)


def test_get_end_timestamp_attrs():
    series = pd.Series()
    series.attrs = {
        "single_metric_dataset_metadata": {
            "ref_interval_end_timestamp": "2025-05-28T18:00:00+02:00"
        }
    }
    timestamp = _get_end_timestamp(series, None)
    assert isinstance(timestamp, pd.Timestamp)


def test_get_end_timestamp_plot_target_settings():
    plot_target_settings_mock = MagicMock(
        return_value=PlotTargetSettings(datetime_x_axes_range_end="2025-05-28T18:00:00+02:00")
    )
    with patch("hdhelpers.helper_functions.get_plot_target_settings", plot_target_settings_mock):
        timestamp = _get_end_timestamp(pd.Series(), None)
        assert isinstance(timestamp, pd.Timestamp)


def test_get_end_none():
    timestamp = _get_end_timestamp(pd.Series(), None)
    assert timestamp is None


def test_get_start_none():
    timestamp = _get_start_timestamp(pd.Series(), None)
    assert timestamp is None


def test_pad_start():
    start = pd.to_datetime("2025-05-28T09:00:00+02:00")
    padded_start = _pad_start(start, "1h")
    assert isinstance(padded_start, pd.Timestamp)
    assert padded_start < start


def test_pad_end():
    end = pd.to_datetime("2025-05-28T18:00:00+02:00")
    padded_end = _pad_end(end, "1h")
    assert isinstance(padded_end, pd.Timestamp)
    assert padded_end > end


def test_pad_start_wrong_padding():
    timestamp = pd.to_datetime("2025-05-28T09:00:00+02:00")
    with pytest.raises(HelperException):
        _pad_start(timestamp, "foo")


def test_pad_end_wrong_padding():
    timestamp = pd.to_datetime("2025-05-28T09:00:00+02:00")
    with pytest.raises(HelperException):
        _pad_end(timestamp, "foo")


def test_to_datetime_int():
    timestamp = 1748415600
    timestamp = _to_datetime(timestamp)
    assert isinstance(timestamp, pd.Timestamp)


def test_to_datetime_str():
    timestamp = "2025-05-28T09:00:00+02:00"
    timestamp = _to_datetime(timestamp)
    assert isinstance(timestamp, pd.Timestamp)


def test_to_datetime_none():
    timestamp = None
    timestamp = _to_datetime(timestamp)
    assert timestamp is None


def test_to_datetime_float():
    timestamp = 3.14
    with pytest.raises(HelperException):
        timestamp = _to_datetime(timestamp)
